# HyTeG Operator Generator
# Copyright (C) 2024  HyTeG Team
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from hog.recipes.common import *


def integrand(
    *,
    jac_a_inv,
    jac_a_abs_det,
    jac_b_inv,
    jac_b_abs_det,
    u,
    v,
    grad_u,
    grad_v,
    k,
    x,
    scalars,
    tabulate,
    volume_geometry,
    **_,
):

    grad_u_chain = jac_b_inv.T * tabulate(jac_a_inv.T * grad_u)
    grad_v_chain = jac_b_inv.T * tabulate(jac_a_inv.T * grad_v)

    def symm_grad(w):
        return 0.5 * (w + w.T)

    symm_grad_u = symm_grad(grad_u_chain)
    symm_grad_v = symm_grad(grad_v_chain)

    div_u = (jac_b_inv.T * tabulate(jac_a_inv.T * grad_u)).trace()
    div_v = (jac_b_inv.T * tabulate(jac_a_inv.T * grad_v)).trace()

    c_rot_penalty = scalars("c_rot_penalty")
    
    if volume_geometry.dimensions == 2:
        rigid_z = sp.Matrix([[-x[1, 0]], [x[0, 0]]])
        rigid_body_model_val = dot(rigid_z, u) * dot(rigid_z, v)
    elif volume_geometry.dimensions == 3:
        rigid_z = sp.Matrix([[-x[1, 0]], [x[0, 0]], [0]])
        rigid_z = rigid_z / rigid_z.norm()

        rigid_y = sp.Matrix([[0], [-x[2, 0]], [x[1, 0]]])
        rigid_y = rigid_y / rigid_y.norm()

        rigid_x = sp.Matrix([[x[2, 0]], [0], [-x[0, 0]]])
        rigid_x = rigid_x / rigid_x.norm()

        rigid_body_model_val = dot(rigid_z, u) * dot(rigid_z, v) + dot(rigid_y, u) * dot(rigid_y, v) + dot(rigid_x, u) * dot(rigid_x, v)

    return k["mu"] * (
        (
            double_contraction(2 * symm_grad_u, symm_grad_v)
            * tabulate(jac_a_abs_det)
            * jac_b_abs_det
        )
        - (2.0 / 3.0) * div_u * div_v * tabulate(jac_a_abs_det) * jac_b_abs_det
    ) + (c_rot_penalty * rigid_body_model_val) * tabulate(jac_a_abs_det) * jac_b_abs_det
